# -*- coding: utf-8 -*-
#   Copyright (C) 2020 Rocky Bernstein
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from trepan.processor.command.base_subcmd import DebuggerSubcommand
from trepan.lib.complete import complete_token


class SetLogLevel(DebuggerSubcommand):
    """**set loglevel** [ **on** | **off** | **debug** | **info** | **critical* | *warn* ]

Show loglevel PyVM logger messages. Initially logtracing is `off`.

However running `set loglevel` will turn it on and set the log level to `debug`. So it's the
same thing as `set loglevel debug`.

If you want the less verbose messages, use `info`. And to turn off,
(except critical errors), use `off`.

Examples:
---------

     set loglevel         # turns x-python on info logging messages
     set loglevel info    # same as above
     set loglevel debug   # turn on info and debug logging messages
     set loglevel off     # turn off all logging messages except critical ones
    """

    in_list    = True
    min_abbrev = len('logl')

    logger_choices = frozenset(["debug", "info", "warn", "warning", "critical", "off", "on"])

    def complete(self, prefix):
        return complete_token(SetLogLevel.logger_choices, prefix)

    def get_loglevel_level(self, arg):
        if not arg: return "info"
        if arg in SetLogLevel.logger_choices:
            return arg
        else:
            self.errmsg('Expecting %s"; got %s' %
                        (', '.join(SetLogLevel.logger_choices), arg))
            return None
        pass


    def run(self, args):
        if len(args) == 0:
            loglevel_level = logging.INFO
        else:
            level_str = self.get_loglevel_level(args[0])
            if not level_str:
                return
            if level_str in ("off", "critical"):
                loglevel_level = logging.CRITICAL
            elif level_str in ("info", "on") :
                loglevel_level = logging.INFO
            elif level_str in ("warn", "warning") :
                loglevel_level = logging.WARNING
            else:
                assert level_str == "debug"
                loglevel_level = logging.DEBUG
            pass

        # Remove all handlers associated with the root logger object.
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)
            pass
        logging.basicConfig(level=loglevel_level)
        self.proc.commands["show"].run(["show", "loglevel"])
        return

    pass

if __name__ == '__main__':
    from trepan.processor.command.set_subcmd.__demo_helper__ import demo_run
    demo_run(SetLogLevel, [])
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2020 Rocky Bernstein
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#    02110-1301 USA.

# Helper function for Processor. Put here so we
# can use this in a couple of processors.

from typing import Any, Optional, Callable

from trepanxpy.fmt import format_instruction_with_highlight

ALL_EVENT_NAMES = (
    "c_call",
    "c_exception",
    "c_return",
    "call",
    "exception",
    "line",
    "return",
    "instruction",
    "yield",
)

# If you want short strings for the above event names
EVENT2SHORT = {
    "c_call": "C>",
    "c_exception": "C!",
    "c_return": "C<",
    "call": "->",
    "exception": "!!",
    "line": "--",
    "instruction": "..",
    "return": "<-",
    "yield": "<>",
    "fatal": "XX",
}

ALL_EVENTS = frozenset(ALL_EVENT_NAMES)


class XPyPrintProcessor(object):
    """
    A processor that just prints out events as we see them. This
    is suitable for example for line/call tracing. We assume that the
    caller is going to filter out which events it wants printed or
    whether it wants any printed at all.
    """

    def __init__(self, core_obj: Any, opts=None):
        self.core = core_obj
        self.debugger = core_obj.debugger
        return

    # FIXME: remove byteName and arguments and have the last instruction sent back
    def event_hook(
        self,
        event: str,
        offset: int,
        byteName: str,
        byteCode: int,
        line_number: int,
        intArg: Optional[int],
        event_arg: Any,
        vm: Any,
        prompt="trepan-xpy-trace",
    ) -> Optional[Callable]:
        "A simple event processor that prints out events."
        if vm.frame:
            if offset >= 0:
                print("%-12s - %s" % (event,
                                      format_instruction_with_highlight(
                                          vm.frame,
                                          vm.opc,
                                          byteName,
                                          intArg,
                                          event_arg,
                                          offset,
                                          line_number,
                                          extra_debug=False,
                                          settings=self.debugger.settings,
                                          show_line=True,
                                          vm=vm
                                      )))
            else:
                frame = vm.frame
                lineno = frame.line_number()
                filename = self.core.canonic_filename(frame)
                filename = self.core.filename(filename)
                print("%s - %s:%d" % (event, filename, lineno))
        else:
                print("%s" % (event,))

        return self.event_hook

    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2017, 2020 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect
from xdis import findlinestarts

from trepan.processor.command.base_subcmd import DebuggerSubcommand
from trepan.lib.disassemble import disassemble_bytes

# FIXME: this could be combined with trepan3k's `info pc`, which doesn't
# require a running program but uses use f_lasti.
# What we have here is less desirable the presence of exceptions,
# but this code would have to be looked over to see if it too would
# work.
class InfoPC(DebuggerSubcommand):
    """**info pc**

List the current program counter or bytecode offset,
and disassemble the instructions around that.

See also:
---------

`info line`, `info program`
"""

    min_abbrev = 2  # Need at least info 'pc'
    max_args = 0
    need_stack = True
    short_help = "Show Program Counter or Instruction Offset information"

    def run(self, args):
        """Program counter."""
        proc = self.proc
        curframe = proc.curframe
        if curframe:
            line_no = inspect.getlineno(curframe)
            offset = curframe.f_lasti
            self.msg("PC offset is %d." % offset)
            offset = max(offset, 0)
            code = curframe.f_code
            co_code = code.co_code
            disassemble_bytes(
                self.msg,
                self.msg_nocr,
                code = co_code,
                lasti = offset,
                cur_line = line_no,
                start_line = line_no - 1,
                end_line = line_no + 1,
                varnames=code.co_varnames,
                names=code.co_names,
                constants=code.co_consts,
                cells=code.co_cellvars,
                freevars=code.co_freevars,
                linestarts=dict(findlinestarts(code)),
                end_offset=offset + 10,
                opc=proc.vm.opc,
            )
            pass
        return False

    pass


if __name__ == "__main__":
    from trepan.processor.command import mock, info as Minfo

    d, cp = mock.dbg_setup()
    i = Minfo.InfoCommand(cp)
    sub = InfoPC(i)
    sub.run([])



# -*- coding: utf-8 -*-
#  Copyright (C) 2009, 2013-2015, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os.path as osp, sys

from trepan.processor.command.base_cmd import DebuggerCommand


class ReturnCommand(DebuggerCommand):
    """**return** [*value*]

Cause an immediate return giving *value* as the return value.
If no value is specified, we'll use `None`.

See also:
---------

`step`, `stepi`.
"""

    aliases = ("ret",)
    category = "running"
    execution_set = ["Running"]
    min_args = 0
    max_args = 1
    name = osp.basename(__file__).split(".")[0]
    need_stack = True
    short_help = "Immediate return with value"

    def run(self, args):
        proc = self.proc
        if proc.stack is None:
            return False

        if len(args) == 0:
            proc.vm.return_value = None
        else:
            text = self.proc.current_command[len(self.proc.cmd_name):]
            pass
        text = text.strip()
        try:
            proc.vm.return_value = proc.eval(text)
        except:
            pass

        proc.return_status = "return"
        proc.continue_running = True  # Break out of command read loop
        return True

    pass


if __name__ == "__main__":
    from trepan.processor.command.mock import MockDebugger
    from xpython.vmtrace import PyVMTraced
    d = MockDebugger()
    proc = d.core.processor
    proc.vm = PyVMTraced(None)
    cmd = ReturnCommand(d.core.processor)

    def demo_return(cmd):
        for c in (
            ["return", "1"],
            ["return", "wrong", "number", "of", "args"],
            ["return", "5"],
        ):
            proc.vm.return_status = "Initial"
            proc.continue_running = False
            proc.stack = [(sys._getframe(0), 14,)]
            result = cmd.run(c)
            print("Execute result: %s" % result)
            print(
                "continue_running: %s, return status: %s"
                % (proc.continue_running,
                   proc.vm.return_status)
            )
            pass
        return

    demo_return(cmd)
    pass



from os.path import basename
from numbers import Number as NumberType
from pygments.token import Number, Text

from trepan.lib.format import (
    Filename,
    Function,
    LineNumber,
    Name,
    Opcode,
    Operator,
    String,
    format_token,
)


LINE_NUMBER_WIDTH = 4
LINE_NUMBER_WIDTH_FMT = "%%-%dd" % LINE_NUMBER_WIDTH
LINE_NUMBER_SPACES = " " * (LINE_NUMBER_WIDTH + len("L. ")) + "@"


def format_instruction_with_highlight(
    frame,
    opc,
    byte_name,
    int_arg,
    arguments,
    offset,
    line_number,
    extra_debug,
    settings,
    show_line=True,
    vm=None,
    repr=repr
):
    """A version of x-python's format_instruction() with terminal highlighting"""

    highlight=settings.get("highlight", False)
    code = frame.f_code if frame else None
    byteCode = opc.opmap.get(byte_name, 0)
    if isinstance(arguments, list) and arguments:
        arguments = arguments[0]
    argrepr = arguments

    fmt_type = Text

    if vm and byte_name in vm.byteop.stack_fmt:
        stack_args = vm.byteop.stack_fmt[byte_name](vm, int_arg, repr)
    else:
        stack_args = ""

    if hasattr(opc, "opcode_arg_fmt") and byte_name in opc.opcode_arg_fmt:
        argrepr = opc.opcode_arg_fmt[byte_name](int_arg)
    elif int_arg is None:
        argrepr = ""
    elif byteCode in opc.NAME_OPS | opc.FREE_OPS | opc.LOCAL_OPS:
        fmt_type = Name
    elif byteCode in opc.JREL_OPS | opc.JABS_OPS:
        fmt_type = Number
        argrepr = str(argrepr)
    elif byteCode in opc.COMPARE_OPS:
        fmt_type = Operator
        argrepr = opc.cmp_op[int_arg]
    elif byteCode == opc.LOAD_CONST:
        if isinstance(argrepr, str):
            fmt_type = String
        elif isinstance(argrepr, NumberType):
            fmt_type = Number
        argrepr = repr(argrepr)

    if line_number is None or not show_line:
        line_str = LINE_NUMBER_SPACES
    else:
        number_str = format_token(
            LineNumber, LINE_NUMBER_WIDTH_FMT % line_number, highlight=highlight
        )
        line_str = "L. %s@" % number_str
        format_token(fmt_type, argrepr),

    mess = "%s%3d: %s%s %s" % (
        line_str,
        offset,
        format_token(Opcode, byte_name),
        stack_args,
        format_token(fmt_type, argrepr),
    )
    if extra_debug and frame:
        filename = basename(code.co_filename) if settings["basename"] else code.co_filename
        mess += " %s in %s:%s" % (
            format_token(Function, code.co_name, highlight=highlight),
            format_token(Filename, filename, highlight=highlight),
            format_token(LineNumber, str(frame.f_lineno), highlight=highlight),
        )
    return mess



# -*- coding: utf-8 -*-
#   Copyright (C) 2020 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command.base_subcmd import DebuggerSubcommand


class InfoBlocks(DebuggerSubcommand):
    """**info blocks**

Show the block-unwinding structure of the current frame

See also:
---------

`show stack`,
"""

    min_abbrev = 1  # Need at least info 'b'
    max_args = 0
    need_stack = True
    short_help = "Show the block-unwinding structure of the current frame"

    def run(self, args):
        """Block Stack."""
        frame = self.proc.vm.frame
        if not hasattr(frame, "block_stack"):
            self.errmsg("frame %s doesn't have a block stack" % frame)
        else:
            block_stack = frame.block_stack
            if len(block_stack) == 0:
                self.msg("Block stack is empty")
            else:
                for i, obj in enumerate(reversed(block_stack)):
                    self.msg("%2d: %s" % (i, obj))
        return False

    pass


# if __name__ == "__main__":
#     from trepan.processor.command import mock, info as Minfo

#     d, cp = mock.dbg_setup()
#     i = Minfo.InfoCommand(cp)
#     sub = InfoBlocks(i)
#     sub.run([])



# -*- coding: utf-8 -*-
#  Copyright (C) 2009, 2013, 2015, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os.path as osp

from xpython.vmtrace import PyVMEVENT_ALL

from trepan.processor.command.base_cmd import DebuggerCommand

# Our local modules

class StepICommand(DebuggerCommand):
    """**stepi** [*count*]

Step an instruction, stopping at the next instruction or event.

With an integer argument, step that many times.


Examples:
---------

  stepi        # step 1 event, *any* event
  stepi 1      # same as above
  stepi 5/5+0  # same as above

Related and similar is the `next` command. More general is the `step` command

See also:
---------

`step` `next`, `skip`, `jump` (there's no `hop` yet), `continue`, `return` and
`finish` for other ways to progress execution.
"""

    aliases       = ('si',)
    category      = 'running'
    min_args      = 0
    max_args      = None
    execution_set = ['Running']
    name          = osp.basename(__file__).split('.')[0]
    need_stack    = True
    short_help    = 'Step instruction (possibly entering called functions)'

    def run(self, args):
        proc = self.proc
        core = self.core
        if len(args) <= 1:
            core.step_ignore = 0
        else:
            pos = 1
            if pos == len(args) - 1:
                core.step_ignore = proc.get_int(args[pos], default=1,
                                                          cmdname='step')
                if core.step_ignore is None: return False
                # 0 means stop now or step 1, so we subtract 1.
                core.step_ignore -= 1
                pass
            elif pos != len(args):
                self.errmsg("Invalid additional parameters %s"
                            % ' '.join(args[pos]))
                return False
            pass

        proc.vm.frame.event_flags = PyVMEVENT_ALL

        core.different_line   = True # Mcmdfns.want_different_line(args[0], self.settings['different'])
        core.stop_level       = None
        core.last_frame       = None
        core.stop_on_finish   = False
        proc.continue_running = True  # Break out of command read loop
        return True
    pass

if __name__ == '__main__':
    from trepan.processor.command.mock import MockDebugger
    d = MockDebugger()
    cmd = StepICommand(d.core.processor)
    for c in (['si', '5'],
              ['stepi', '1+2'],
              ['si', 'foo']):
        d.core.step_ignore = 0
        cmd.proc.continue_running = False
        result = cmd.run(c)
        print('Execute result: %s' % result)
        print('step_ignore %s' % repr(d.core.step_ignore))
        print('continue_running: %s' % cmd.proc.continue_running)
        pass
    # for c in (['si'], ['stepi']):
    #     d.core.step_ignore = 0
    #     cmd.continue_running = False
    #     result = cmd.run(c)
    #     print('different line %s:' % c[0], cmd.core.different_line)
    #     pass
    pass



# -*- coding: utf-8 -*-
#  Copyright (C) 2009, 2015, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from xpython.vm import log
from trepan.processor.command.base_subcmd import DebuggerSubcommand

LOGLEVEL2NAME = {
    50: "critical",
    40: "error",
    30: "warning",
    20: "info",
    10: "debug",
    00: "notset"
}


class ShowLogLevel(DebuggerSubcommand):
    """**show loglevel**

Show VM logger log level setting.

See also:
---------

`set logtrace`"""

    min_abbrev = len("logl")
    short_help = "Show VM logger log level setting"
    pass

    def run(self, args):
        if len(args) != 0:
            self.errmsg("Expecting no args")
            return

        log_str = LOGLEVEL2NAME.get(log.getEffectiveLevel(), "unknown")
        self.msg("PyVM log level is %s" % log_str)
        return
    pass

if __name__ == "__main__":
    # from trepan.processor.command.show_subcmd.__demo_helper__ import demo_run

    # demo_run(ShowLogLevel)
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2020 Rocky Bernstein
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

from getopt import getopt, GetoptError

from trepan.processor.command.base_subcmd import DebuggerSubcommand
from trepan.processor.cmdfns import get_an_int
from trepan.misc import wrapped_lines


class SetPC(DebuggerSubcommand):
    """**set pc** [ *flag* ] *line-or-offset*

Set the PC to run at *line-or-offset*

*flag* can be one of:

* `--offset` | `-o` to indicate an offset
* `--line` | `-l` to indicate a line number
* `=` which is ignored. It is syntactic sugar

Alternatively, the line-or-offset can be indicated by prefixing the
number with `l` or `@` respectively.

Examples
--------

    set pc 0
    set pc @0            # same as above
    set pc --offset 0    # same as above
    set pc -o 0          # same as above
    set pc = 0           # same as above

    set pc l2            # same as above
    set pc --line 2      # same as above
    set pc -l 2          # same as above
    set pc = l0          # same as above

See Also
--------

`info pc`
"""

    in_list    = True
    min_abbrev = len('pc')
    max_args = 2
    need_stack = True
    short_help = "Set Program Counter"

    def run(self, args):
        mainfile = self.core.filename(None)
        if self.core.is_running():
            proc = self.proc

            is_offset = True
            n = len(args)
            if n == 2:
                if args[0] == "=":
                    args[0] = args[1]
                else:
                    try:
                        opts, args = getopt(args, "hlo",
                                            ["help", "line", "offset"])
                    except GetoptError as err:
                        # print help information and exit:
                        print(str(err))  # will print something like "option -a not recognized"
                        return

                    for o, a in opts:
                        if o in ("-h", "--help"):
                            proc.commands["help"].run(["help", "set", "pc"])
                            return
                        elif o in ("-l", "--line", ):
                            is_offset = False
                        elif o in ("-o", "--offset"):
                            is_offset = True
                        else:
                            self.errmsg("unhandled option '%s'" % o)
                        pass
                    pass
                pass

            frame = proc.frame
            if frame:
                code = frame.f_code
                if is_offset:
                    min_value = 0
                    max_value = len(code.co_code) - 1
                elif len(code.co_lnotab) > 0:
                    min_value = code.co_firstlineno
                    # FIXME: put this in a function in xdis somewhere.
                    # Also: won't need ord() in Python 2.x
                    # max_value = min_value + sum([ord(i) for i in code.co_lnotab[1::2]])
                    max_value = min_value + sum(code.co_lnotab[1::2])
                else:
                    min_value = max_value = code.co_firstline

                offset = get_an_int(
                    self.errmsg,
                    args[0],
                    "The 'pc' command requires an integer offset or a line.",
                    min_value, max_value,
                )
                if offset is None:
                    return None
                # FIXME: We check that offset points to an
                # opcode or a valid line as opposed to an operand
                if not is_offset:
                    self.errmsg("Sorry, line numbers not handled right now.")
                    return None

                self.list_offset = frame.f_lasti = offset
                frame.fallthrough = False
                self.list_lineno = frame.f_lineno = frame.line_number()
                self.msg("Execution set to resume at offset %d" % offset)
                self.return_status = "skip"
                return None
            else:
                self.errmsg("Oddly, a frame is not set - nothing done.")
                pass
        else:
            if mainfile:
                part1 = "Python program '%s'" % mainfile
                msg = "is not currently running. "
                self.msg(wrapped_lines(part1, msg, self.settings["width"]))
            else:
                self.msg("No Python program is currently running.")
                pass
            self.msg(self.core.execution_status)
            pass
        return None

    pass

if __name__ == '__main__':
    from trepan.processor.command.set_subcmd.__demo_helper__ import demo_run

    demo_run(SetPC, [0])
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2020 Rocky Bernstein <rocky@gnu.org>

from typing import List, Dict, Any
import os
import sys
from xpython.execfile import run_python_file, run_python_string, NoSourceError

# Default settings used here
from trepanxpy.debugger_defaults import DEBUGGER_SETTINGS
from trepan.interfaces.user import UserInterface
import trepan.misc as Mmisc
from trepan.exception import DebuggerQuit, DebuggerRestart

from trepanxpy.core import TrepanXPyCore
from trepanxpy.fmt import format_instruction_with_highlight
from trepanxpy.processor.cmd import XPyCommandProcessor
from trepanxpy.processor.trace import XPyPrintProcessor


class Debugger(object):
    def __init__(
        self, string_or_path: str, is_file: bool, trace_only: bool, args: List[str]
    ):
        """Create a debugger object. But depending on the value of
        key 'start' inside hash 'opts', we may or may not initially
        start debugging.

        See also Debugger.start and Debugger.stop.
        """

        def instruction_fmt_func(frame, opc, byte_name, int_arg, arguments, offset, line_number,
                                 extra_debug, vm):
            return format_instruction_with_highlight(
                frame=frame,
                opc=opc,
                byte_name=byte_name,
                int_arg = int_arg,
                arguments = arguments,
                offset = offset,
                line_number = line_number,
                extra_debug = extra_debug,
                settings=self.settings,
                vm = vm,
                show_line = True
            )

        self.mainpyfile = None
        self.thread = None

        completer = lambda text, state: self.complete(text, state)
        interface_opts = {
            "complete": completer,
            "debugger_name": "trepan-xpy",
        }
        interface = UserInterface(opts=interface_opts)
        self.intf = [interface]

        # main_dirname is the directory where the script resides.
        # Filenames in co_filename are often relative to this.
        self.main_dirname = os.curdir

        self.filename_cache: Dict[str, Any] = {}
        self.settings = DEBUGGER_SETTINGS
        self.core = TrepanXPyCore(self, {})
        if trace_only:
            processor = XPyPrintProcessor(self.core)
        else:
            processor = XPyCommandProcessor(self.core)
        self.callback_hook = processor.event_hook

        # Save information for restarting
        self.program_sys_argv = None
        self.orig_sys_argv = list(sys.argv)

        if is_file:
            mainpyfile = self.core.canonic(string_or_path)
            run_fn = run_python_file
        else:
            mainpyfile = string_or_path
            run_fn = run_python_string

        while True:
            print("Running x-python %s with %s" % (string_or_path, args))
            try:
                run_fn(
                    string_or_path,
                    args,
                    callback=self.callback_hook,
                    format_instruction=instruction_fmt_func,
                )
            except DebuggerQuit:
                break
            except DebuggerRestart:
                self.core.execution_status = "Restart requested"
                if self.program_sys_argv:
                    sys.argv = list(self.program_sys_argv)
                    part1 = "Restarting %s with arguments:" % self.core.filename(
                        mainpyfile
                    )
                    args = " ".join(self.program_sys_argv[1:])
                    self.intf[-1].msg(
                        Mmisc.wrapped_lines(part1, args, self.settings["width"])
                    )
                else:
                    break
            except (FileNotFoundError, NoSourceError) as e:
                self.intf[-1].msg(str(e))
                sys.exit(1)
            except SystemExit:
                # In most cases SystemExit does not warrant a post-mortem session.
                break
            else:
                if trace_only:
                    break
                msg = "The program finished - press enter to restart; anything else terminates. ? "
                response = input(msg)
                if response != "":
                    break
                pass

    def restart_argv(self):
        """Return an array that would be execv-ed  to restart the program"""
        return self.orig_sys_argv or self.program_sys_argv



# -*- coding: utf-8 -*-
# This file is stripped down from the one in trepan3k
#
#   Copyright (C) 2008-2010, 2013-2020 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
import linecache, sys, re, inspect
import importlib
import pyficache
import os.path as osp

from typing import Any, Optional, Set

# Note: the module name pre 3.2 is repr
from reprlib import Repr

from pygments.console import colorize

import trepan.lib.bytecode as Mbytecode
import trepan.lib.display as Mdisplay
import trepan.misc as Mmisc
import trepan.lib.thred as Mthread
import trepan.processor.complete as Mcomplete

from trepan.processor.cmdproc import (
    CommandProcessor,
    arg_split,
    resolve_name,
    print_location,
)

from trepanxpy.processor.trace import EVENT2SHORT
from trepanxpy.fmt import format_instruction_with_highlight


warned_file_mismatches: Set[str] = set()


def get_srcdir():
    filename = osp.normcase(osp.dirname(osp.abspath(__file__)))
    return osp.realpath(filename)


# Default settings for command processor method call
DEFAULT_PROC_OPTS = {
    # A list of debugger initialization files to read on first command
    # loop entry.  Often this something like [~/.config/trepan-xp/profile] which the
    # front-end sets.
    "initfile_list": []
}

# FIXME: can't figure out how to delegate specific fns so we'll subclass instead.
class XPyCommandProcessor(CommandProcessor):
    def __init__(self, core_obj, opts=None):
        super().__init__(core_obj, opts)
        self.core = core_obj
        self.debugger = core_obj.debugger

        self.continue_running = False  # True if we should leave command loop
        self.event2short = dict(EVENT2SHORT)
        self.event2short["signal"] = "?!"
        self.event2short["brkpt"] = "xx"

        self.optional_modules = ()

        # command argument string. Is like current_command, but the part
        # after cmd_name has been removed.
        self.cmd_argstr = ""

        # command name before alias or macro resolution
        self.cmd_name = ""
        self.cmd_queue = []  # Queued debugger commands
        self.completer = lambda text, state: Mcomplete.completer(self, text, state)
        self.current_command = ""  # Current command getting run
        self.debug_nest = 1
        self.display_mgr = Mdisplay.DisplayMgr()
        self.intf = core_obj.debugger.intf
        self.last_command = None  # Initially a no-op
        self.precmd_hooks = []

        # FIXME: can we adjust this to also show the instruction?
        self.location = lambda: self

        self.preloop_hooks = []
        self.postcmd_hooks = []

        # Note: prompt_str's value set below isn't used. It is
        # computed dynamically. The value is suggestive of what it
        # looks like.
        self.prompt_str = "(trepan-xpy) "

        # Stop only if line/file is different from last time
        self.different_line = None

        # These values updated on entry. Set initial values.
        self.curframe = None
        self.event = None
        self.event_arg = None
        self.frame = None
        self.list_lineno = 0  # last list number used in "list"
        self.list_offset = -1  # last list number used in "disassemble"
        self.list_obj = None
        self.list_filename = None  # last filename used in list
        self.list_orig_lineno = 0  # line number of frame or exception on setup
        self.list_filename = None  # filename of frame or exception on setup

        self.macros = {}  # Debugger Macros

        # Create a custom safe Repr instance and increase its maxstring.
        # The default of 30 truncates error messages too easily.
        self._repr = Repr()
        self._repr.maxstring = 100
        self._repr.maxother = 60
        self._repr.maxset = 10
        self._repr.maxfrozen = 10
        self._repr.array = 10
        self.stack = []
        self.thread_name = None
        self.frame_thread_name = None

        get_option = lambda key: Mmisc.option_set(opts, key, DEFAULT_PROC_OPTS)
        initfile_list = get_option("initfile_list")
        for init_cmdfile in initfile_list:
            self.queue_startfile(init_cmdfile)

        # FIXME: This doesn't work
        # # Delegate functions here:
        # self.cmdproc = CommandProcessor(self)
        # for method in (
        #         "_saferepr",
        #         "add_preloop_hook",
        #         "defaultFile",
        #         "eval",
        #         "exec_line",
        #         "forget",
        #         "get_an_int",
        #         "get_int_noerr",
        #         "getval",
        #         "ok_for_running",
        #         "process_commands",
        #         "queue_startfile",
        #         "remove_preloop_hook",
        #         "setup",
        #         "undefined_cmd",
        #         "read_history_file",
        #         "write_history_file",
        #         ):
        #     setattr(self, method, getattr(cmdproc, method))

        # Remove trepan3k commands which aren't valid here, and those specific to trepan-xpy
        remove_commands = (
            "break",
            "continue",
            "finish",
            "next",
            "quit",
            "set",
            "step",
            "tbreak",
        )
        new_instances = []
        for cmd in self.cmd_instances:
            if cmd.name in remove_commands:
                del self.commands[cmd.name]
            else:
                new_instances.append(cmd)
                pass
            pass
        self.cmd_instances = new_instances

        new_commands = self._update_commands()
        for new_command in new_commands:
            self.commands[new_command.name] = new_command
        self.cmd_instances += new_commands
        self._populate_cmd_lists()

        if self.debugger.settings["autopc"]:
            self.commands["set"].run(["set", "autopc"])
        return

    def update_commands_easy_install(self, Mcommand):
        """
        Add files in filesystem to self.commands.
        If running from source or from an easy_install'd package, this is used.
        """
        cmd_instances = []

        for mod_name in Mcommand.__modules__:
            if mod_name in ("info_sub", "set_sub", "show_sub",):
                pass
            import_name = "%s.%s" % (Mcommand.__name__, mod_name)
            try:
                command_mod = importlib.import_module(import_name)
            except:
                if mod_name not in self.optional_modules:
                    print("Error importing %s: %s" % (mod_name, sys.exc_info()[0]))
                    pass
                continue

            classnames = [
                tup[0]
                for tup in inspect.getmembers(command_mod, inspect.isclass)
                if (
                    tup[0] != "DebuggerCommand"
                    and not tup[0].startswith("Trepan3k")
                    and tup[0].endswith("Command")
                )
            ]
            for classname in classnames:
                if False:
                    instance = getattr(command_mod, classname)(self)
                    cmd_instances.append(instance)
                else:
                    try:
                        instance = getattr(command_mod, classname)(self)
                        cmd_instances.append(instance)
                    except:
                        print(
                            "Error loading %s from %s: %s"
                            % (classname, mod_name, sys.exc_info()[0])
                        )
                        pass
                    pass
                pass
            pass
        return cmd_instances

    def set_prompt(self, prompt="trepan-xpy"):
        if self.thread_name and self.thread_name != "MainThread":
            prompt += ":" + self.thread_name
            pass
        self.prompt_str = "%s%s%s" % (
            "(" * self.debug_nest,
            prompt,
            ")" * self.debug_nest,
        )
        highlight = self.debugger.settings["highlight"]
        if highlight and highlight in ("light", "dark"):
            self.prompt_str = colorize("underline", self.prompt_str)
        self.prompt_str += " "

    def event_hook(
        self,
        event: str,
        offset: int,
        byteName: str,
        byteCode: int,
        line_number: int,
        intArg: Optional[int],
        event_arg: Any,
        vm: Any,
        prompt="trepan-xpy",
    ):
        "command event processor: reading a commands do something with them."

        def frame_setup(frame):
            filename = frame.f_code.co_filename
            lineno = frame.f_lineno
            line = linecache.getline(filename, lineno, frame.f_globals)
            if not line:
                opts = {
                    "output": "plain",
                    "reload_on_change": self.settings("reload"),
                    "strip_nl": False,
                }
                m = re.search("^<frozen (.*)>", filename)
                if m and m.group(1):
                    filename = pyficache.unmap_file(m.group(1))
                line = pyficache.getline(filename, lineno, opts)
            self.current_source_text = line
            return line, filename

        self.vm = vm
        self.frame = vm.frame
        self.event = event
        self.event_arg = event_arg

        # In order to follow Python's sys.settrace()'s convention:
        # returning "None" turns off tracing for the scope.
        # However we do not need to return a reference to ourself,
        # a callable (this may be allowed in the future though).
        # Instead for now a string status is returned
        # * "skip" for skip next instruction, and
        # * "return" for immediate return
        # * "finish" for "step out"

        self.return_status = True

        if event == "fatal":
            self.core.execution_status = "Terminated"
            # One last hurrah!

            tb = vm.last_traceback
            if tb:
                frame_setup(tb.tb_frame)
                self.vm.frames = []
                while tb:
                    self.vm.frames.insert(0, tb.tb_frame)
                    tb = tb.tb_next
                self.curframe = self.frame = self.vm.frames[0]
                self.setup()
                self.curindex = len(vm.frames) - 1
                print_location(self)

            self.set_prompt("trepan-xpy:pm")
            self.process_commands()
            return None

        if self.vm.frame:
            self.core.execution_status = "Running"
        else:
            self.core.execution_status = "Terminated"
            return

        line, filename = frame_setup(self.frame)
        if self.settings("skip"):
            # Note that in contrast to skipping intructions
            # when return_status is set to "skip", here
            # we are execution the instruction but just skipping
            # any handling this instruction the debugger.
            if Mbytecode.is_def_stmt(line, self.frame):
                return self
            if Mbytecode.is_class_def(line, self.frame):
                return
            pass
        self.thread_name = Mthread.current_thread_name()
        self.frame_thread_name = self.thread_name

        self.setup()
        print_location(self)
        if offset >= 0 and event not in ('call', 'return'):
            self.msg(
                "%s"
                % format_instruction_with_highlight(
                    vm.frame,
                    vm.opc,
                    byteName,
                    intArg,
                    event_arg,
                    offset,
                    line_number,
                    extra_debug=False,
                    settings=self.debugger.settings,
                    show_line=False,  # We show the line number in our location reporting
                    vm=self.vm,
                    repr=self._repr.repr
                )
            )

        self.set_prompt(prompt)
        self.process_commands()
        if filename == "<string>":
            pyficache.remove_remap_file("<string>")
        return self.return_status

    def _update_commands(self):
        """ Create an instance of each of the debugger
        commands. Commands are found by importing files in the
        directory 'command'. Some files are excluded via an array set
        in __init__.  For each of the remaining files, we import them
        and scan for class names inside those files and for each class
        name, we will create an instance of that class. The set of
        DebuggerCommand class instances form set of possible debugger
        commands."""
        from trepanxpy.processor import command as Mcommand

        if hasattr(Mcommand, "__modules__"):
            # This is also used when installing from source
            return self.update_commands_easy_install(Mcommand)
        else:
            return self.populate_commands_pip(Mcommand, "trepanxpy")

    def setup(self):
        """Initialization done before entering the debugger-command
        loop. In particular we set up the call stack used for local
        variable lookup and frame/up/down commands.

        We return True if we should NOT enter the debugger-command
        loop."""
        self.forget()
        self.curindex = 0
        if self.vm.frames:
            stack = self.vm.frames
            if stack == []:
                # FIXME: Just starting up - should there be an event for this
                # Or do we need to do this for all call events?
                stack = [self.frame]
                pass
            self.stack = [(frame, frame.line_number()) for frame in reversed(stack)]
            self.curframe = self.frame
            self.thread_name = "MainThread"
        else:
            self.stack = self.curframe = self.botframe = None
            pass
        if self.curframe:
            self.list_lineno = self.frame.line_number()
            self.list_offset = self.curframe.f_lasti
            self.list_filename = self.curframe.f_code.co_filename
            self.list_object = self.curframe
        else:
            self.list_object = None
            pass
        # if self.execRcLines()==1: return True

        # FIXME:  do we want to save self.list_lineno a second place
        # so that we can do 'list .' and go back to the first place we listed?
        return False


# Demo it
if __name__ == "__main__":
    from trepan.processor.command import mock as Mmock

    d = Mmock.MockDebugger()
    cmdproc = CommandProcessor(d.core)
    print("commands:")
    commands = list(cmdproc.commands.keys())
    commands.sort()
    print(commands)
    print("aliases:")
    aliases = list(cmdproc.aliases.keys())
    aliases.sort()
    print(aliases)
    print(resolve_name(cmdproc, "quit"))
    print(resolve_name(cmdproc, "q"))
    print(resolve_name(cmdproc, "info"))
    print(resolve_name(cmdproc, "i"))
    # print '-' * 10
    # print_source_line(sys.stdout.write, 100, 'source_line_test.py')
    # print '-' * 10
    cmdproc.frame = sys._getframe()
    cmdproc.setup()
    print()
    print("-" * 10)
    cmdproc.location()
    print("-" * 10)
    print(cmdproc.eval("1+2"))
    print(cmdproc.eval("len(aliases)"))
    import pprint

    print(pprint.pformat(cmdproc.category))
    print(arg_split("Now is the time"))
    print(arg_split("Now is the time ;;"))
    print(arg_split("Now is 'the time'"))
    print(arg_split("Now is the time ;; for all good men"))
    print(arg_split("Now is the time ';;' for all good men"))

    print(cmdproc.commands)
    fn = cmdproc.commands["quit"]

    print("Removing non-existing quit hook: %s" % cmdproc.remove_preloop_hook(fn))
    cmdproc.add_preloop_hook(fn)
    print(cmdproc.preloop_hooks)
    print("Removed existing quit hook: %s" % cmdproc.remove_preloop_hook(fn))
    pass



# -*- coding: utf-8 -*-
#  Copyright (C) 2009, 2013, 2015, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os.path as osp

from xpython.vmtrace import (
    PyVMEVENT_LINE,
    PyVMEVENT_CALL,
    PyVMEVENT_RETURN,
    PyVMEVENT_EXCEPTION,
    PyVMEVENT_YIELD
    )

from trepan.processor.command.base_cmd import DebuggerCommand

# Our local modules

class FinishCommand(DebuggerCommand):
    """**finish** [*level*]

Continue execution until leaving the current function. When *level* is
specified, that many frame levels need to be popped. Note that *yield*
and exceptions raised my reduce the number of stack frames. Also, if a
thread is switched, we stop ignoring levels.

See the `break` command if you want to stop at a particular point in a

See also:
---------

`continue`, `step`, `next`.
"""

    aliases = ("fin",)
    category = "running"
    execution_set = ["Running"]
    min_args = 0
    max_args = 1
    name = osp.basename(__file__).split(".")[0]
    need_stack = True
    short_help = "Execute until selected stack frame returns"

    def run(self, args):
        proc = self.proc
        core = self.core
        if len(args) <= 1:
            core.step_ignore = 0
        else:
            pos = 1
            if pos == len(args) - 1:
                core.step_ignore = self.proc.get_int(args[pos], default=1,
                                                          cmdname='step')
                if self.core.step_ignore is None: return False
                # 0 means stop now or step 1, so we subtract 1.
                core.step_ignore -= 1
                pass
            elif pos != len(args):
                self.errmsg("Invalid additional parameters %s"
                            % ' '.join(args[pos]))
                return False
            pass

        core.stop_level       = None
        core.last_frame       = None
        core.stop_on_finish   = False
        proc.continue_running = True  # Break out of command read loop
        proc.return_status = "finish"  # Tell interpreter to step out.
        return True
    pass

if __name__ == '__main__':
    from trepan.processor.command.mock import MockDebugger

    d = MockDebugger()
    cmd = FinishCommand(d.core.processor)
    for c in (["n", "5"], ["next", "1+2"], ["n", "foo"]):
        d.core.step_ignore = 0
        cmd.continue_running = False
        result = cmd.run(c)
        print("Run result: %s" % result)
        print(
            "step_ignore %d, continue_running: %s"
            % (d.core.step_ignore, cmd.continue_running,)
        )
        pass
    for c in (["n"], ["next+"], ["n-"]):
        d.core.step_ignore = 0
        cmd.continue_running = False
        result = cmd.run(c)
        print(cmd.core.different_line)
        pass
    pass



# -*- coding: utf-8 -*-
#  Copyright (C) 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os.path as osp

from trepan.processor.command.show import ShowCommand as Trepan3kShowCommand

class ShowCommand(Trepan3kShowCommand):
    """**show** *subcommand*

Generic command for showing things about the debugger.  You can
give unique prefix of the name of a subcommand to get information
about just that subcommand.

Type `show` for a list of *show* subcommands and what they do.
Type `help show *` for just a list of *show* subcommands.
"""

    category = "status"
    min_args = 0
    max_args = None
    name = osp.basename(__file__).split(".")[0]
    need_stack = False
    short_help = "Show parts of the debugger environment"

    def __init__(self, proc, name="show"):
        """Initialize show subcommands. Note: instance variable name
        has to be setcmds ('set' + 'cmds') for subcommand completion
        to work."""

        super().__init__(proc, name)
        new_cmdlist = []
        for subname in self.cmds.cmdlist:
            if subname in ("dbg_trepan", "events"):
                del self.cmds.subcmds[subname]
            else:
                new_cmdlist.append(subname)
        self.cmds.cmdlist = new_cmdlist
        self._load_debugger_subcommands(name, "trepanxpy")


if __name__ == "__main__":
    from trepan.processor.command import mock

    d, cp = mock.dbg_setup()
    command = ShowCommand(cp, "show")
    command.run(["show"])
    pass



# -*- coding: utf-8 -*-
#  Copyright (C) 2009-2010, 2013, 2015, 2017, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import code, os.path as osp, sys

# Our local modules
from trepan.processor.command.python import (
    PythonCommand as Trepan3kShell,
    interact,
    runcode,
)
from trepan.interfaces.server import ServerInterface


class PythonCommand(Trepan3kShell):
    """**python**

Run Python as a command subshell. The *sys.ps1* prompt will be set to
`trepan-xpy >>> `.

To issue a debugger command use function *dbgr()*. For example:

  dbgr('info program')

The current call frame and Interpreter VM access is available through variables: frame, and vm.
"""

    aliases = ("py", "shell")
    category = "support"
    min_args = 0
    max_args = 1
    name = osp.basename(__file__).split(".")[0]
    need_stack = False
    short_help = "Run a Python interactive shell"

    def run(self, args):
        # See if python's code module is around

        # Python does it's own history thing.
        # Make sure it doesn't damage ours.
        intf = self.debugger.intf[-1]
        if isinstance(intf, ServerInterface):
            self.errmsg("Can't run an interactive shell on a remote session")
            return

        have_line_edit = self.debugger.intf[-1].input.line_edit
        if have_line_edit:
            try:
                self.proc.write_history_file()
            except IOError:
                pass
            pass

        banner_tmpl = """trepan-xpy python shell%s
Use dbgr(*string*) to issue debugger command: *string*"""

        banner_tmpl += "\nVariable 'frame' contains the current frame; 'vm', the current vm object."

        my_locals = {}
        my_globals = None
        proc = self.proc
        if proc.curframe:
            my_globals = proc.curframe.f_globals
            if self.proc.curframe.f_locals:
                my_locals = proc.curframe.f_locals
                pass
            pass

        # Give the user a way to get access frame, vm and debugger commands
        my_locals["vm"] = proc.vm
        my_locals["frame"] = proc.frame
        my_locals["dbgr"] = self.dbgr

        # Change from debugger completion to python completion
        try:
            import readline
        except ImportError:
            pass
        else:
            readline.parse_and_bind("tab: complete")

        sys.ps1 = "trepan-xpy >>> "
        old_sys_excepthook = sys.excepthook
        try:
            sys.excepthook = None
            if len(my_locals):
                interact(
                    banner=(banner_tmpl % " with locals"),
                    my_locals=my_locals,
                    my_globals=my_globals,
                )
            else:
                interact(banner=(banner_tmpl % ""))
                pass
        finally:
            sys.excepthook = old_sys_excepthook

        # restore completion and our history if we can do so.
        if hasattr(self.proc.intf[-1], "complete"):
            try:
                from readline import set_completer, parse_and_bind

                parse_and_bind("tab: complete")
                set_completer(self.proc.intf[-1].complete)
            except ImportError:
                pass
            pass

        if have_line_edit:
            self.proc.read_history_file()
            pass
        return

    pass


if __name__ == "__main__":
    from trepanxpy.debugger import Debugger

    d = Debugger("x + 1", is_file=False, trace_only=False, args=[])
    command = PythonCommand(d.core.processor)
    command.proc.frame = sys._getframe()
    command.proc.setup()
    if len(sys.argv) > 1:
        print("Type Python commands and exit to quit.")
        print(sys.argv[1])
        if sys.argv[1] == "-d":
            print(command.run(["python", "-d"]))
        else:
            print(command.run(["python"]))
            pass
        pass
    pass



"""A main program for trepan-xpy."""
import sys
import click
from typing import List

from trepanxpy.version import VERSION
from trepanxpy.debugger import Debugger

@click.command()
@click.version_option(VERSION, "-V", "--version")
@click.option("-x", "--trace", default=False, required=False, flag_value="trace",
              help="Run with instruction tracing, no interactive debugging (until post-mortem)")
@click.option("-c", "--command-to-run",
              help="program passed in as a string", required=False)
@click.argument("path", nargs=1, type=click.Path(readable=True), required=False)
@click.argument("args", nargs=-1)
def main(trace: bool, path: str, command_to_run: str, args: List[str]):

    # FIXME: This seems to be needed for pyficache to work on relative paths.
    # is this a bug?
    sys.path.append(".")

    is_file = True
    string_or_path = path
    if command_to_run:
        if path or args:
            print("You must pass either a file name or a command string, not both.")
            sys.exit(4)
        string_or_path = command_to_run
        is_file = False
    elif not path:
        print("You must pass either a file name or a command string, neither found.")
        sys.exit(4)

    Debugger(string_or_path, is_file, trace_only=trace, args=args)

if __name__ == "__main__":
    main(auto_envvar_prefix="XPYTHON")



# -*- coding: utf-8 -*-
#  Copyright (C) 2009, 2013, 2015, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os.path as osp

from xpython.vmtrace import (
    PyVMEVENT_ALL,
    PyVMEVENT_INSTRUCTION,
    PyVMEVENT_STEP_OVER,
    )

from trepan.processor.command.base_cmd import DebuggerCommand

# Our local modules

class NextCommand(DebuggerCommand):
    """**next**[**+**|**-**] [*count*]

Execute the current simple statement stopping at the next event but
ignoring steps into function calls at this level,

With an integer argument, perform `next` that many times. However if
an exception occurs at this level, or we *return*, *yield* or the
thread changes, we stop regardless of count.

A suffix of `+` on the command or an alias to the command forces to
move to another line, while a suffix of `-` does the opposite and
disables the requiring a move to a new line. If no suffix is given,
the debugger setting 'different-line' determines this behavior.

See also:
---------

`step`, `skip`, `jump` (there's no `hop` yet), `continue`, and
`finish` for other ways to progress execution.
"""

    aliases = ("next+", "next-", "n", "n-", "n+")
    category = "running"
    execution_set = ["Running"]
    min_args = 0
    max_args = 1
    name = osp.basename(__file__).split(".")[0]
    need_stack = True
    short_help = "Step over"

    def run(self, args):
        proc = self.proc
        core = self.core
        if len(args) <= 1:
            core.step_ignore = 0
        else:
            pos = 1
            if pos == len(args) - 1:
                core.step_ignore = self.proc.get_int(args[pos], default=1,
                                                          cmdname='step')
                if self.core.step_ignore is None: return False
                # 0 means stop now or step 1, so we subtract 1.
                core.step_ignore -= 1
                pass
            elif pos != len(args):
                self.errmsg("Invalid additional parameters %s"
                            % ' '.join(args[pos]))
                return False
            pass

        # Add "step over" to all event flags, but remove instruction events.
        proc.vm.frame.event_flags = (PyVMEVENT_ALL | PyVMEVENT_STEP_OVER) ^ PyVMEVENT_INSTRUCTION

        core.different_line   = True # Mcmdfns.want_different_line(args[0], self.settings['different'])
        core.stop_level       = None
        core.last_frame       = None
        core.stop_on_finish   = False
        proc.continue_running = True  # Break out of command read loop
        return True
    pass

if __name__ == '__main__':
    from trepan.processor.command.mock import MockDebugger

    d = MockDebugger()
    cmd = NextCommand(d.core.processor)
    for c in (["n", "5"], ["next", "1+2"], ["n", "foo"]):
        d.core.step_ignore = 0
        cmd.continue_running = False
        result = cmd.run(c)
        print("Run result: %s" % result)
        print(
            "step_ignore %d, continue_running: %s"
            % (d.core.step_ignore, cmd.continue_running,)
        )
        pass
    for c in (["n"], ["next+"], ["n-"]):
        d.core.step_ignore = 0
        cmd.continue_running = False
        result = cmd.run(c)
        print(cmd.core.different_line)
        pass
    pass
